module Main where

import Integrators
import Physics
import AstroData
--import AsteroidData
--import PlanetData
import AllBodiesData

import System.IO(IOMode(WriteMode), withFile, stderr, hPutStrLn)
import System.Environment(getArgs, getProgName)
import System.Exit(exitFailure)

import Control.Parallel.Strategies(using, rdeepseq, parTraversable)
import Control.DeepSeq(NFData)

{-|
  Default GR cutoff values such that Jupiter is massive enough and Mercury
  close enough for GR effects
-}
defGrcut :: Floating t => CutParams t
defGrcut = CutParams 0.2e-6 0.5

main :: IO ()
main = do args <- getArgs
          case args of
            [steps, dt, filename] -> do
              let (steps', dt') = (1 + (read steps :: Int), read dt :: Double)
              --let ephem = generateEphemerides rkcv8 (majorbodies ++ asteroids) defGrcut dt' steps'
              let ephem = generateEphemeridesPar rkcv8 allbodies defGrcut dt' steps'
              withFile filename WriteMode (\h -> mapM_ (\ob->mapM_ ((hPutStrLn h). formatDoubleOB) ob) ephem)
            _ -> do pn <- getProgName
                    hPutStrLn stderr $ "Usage: " ++ pn ++ " steps dt filename"
                    exitFailure


integrableForceEq :: (Floating f, Ord f) => [ OrbitBody f ] -> CutParams f -> OrbitBody f -> Vec2 f -> Vec2 f
integrableForceEq bodies  grcut (OrbitBody bid' bGM' _ et') svec =
  Vec2 (v svec) $ pNGRFull bodies (OrbitBody bid' bGM' svec et') grcut

stepForwardAll :: (Floating t, Eq t) => Integrator t -> (OrbitBody t -> Vec2 t -> Vec2 t) -> t -> [OrbitBody t] -> [OrbitBody t]
stepForwardAll (SingStepIntegrator i) f dt bodies =
  [ updateSvecs body dt $ i (f body) (brv body) dt | body <- bodies ]
stepForwardAll _ _ _ _ = undefined


stepForwardAllPar :: (Floating t, Eq t, NFData t) => Integrator t -> (OrbitBody t -> Vec2 t -> Vec2 t) -> t -> [ OrbitBody t ] -> [ OrbitBody t ]
stepForwardAllPar (SingStepIntegrator i) f dt bodies =
  fmap method bodies `using` parTraversable rdeepseq
  where
    method body = updateSvecs body dt $ i (f body) (brv body) dt 
stepForwardAllPar _ _ _ _ = undefined

  {-
    -- Silly explicit version of stepForwardAllPar that shows that my current
    -- form is in fact not working, only applying to two cores whereas this
    -- version does succesfully work on three cores. It also sparks way too many
    -- and fails to be any faster.
stepForwardAllPar :: (Floating t, Eq t, NFData t) => Integrator t -> (OrbitBody t -> Vec2 t -> Vec2 t) -> t -> [ OrbitBody t ] -> [ OrbitBody t ]
stepForwardAllPar (SingStepIntegrator i) f dt bodies =
  --fmap method bodies `using` parTraversable rdeepseq
  chunk3 method bodies
  where
    method body = updateSvecs body dt $ i (f body) (brv body) dt 
    chunk3 fn (l1:l2:l3:ls) = (t3tol $ runEval (parTuple3 rdeepseq rdeepseq rdeepseq (fn l1, fn l2, fn l3))) ++ (chunk3 fn ls)
    chunk3 fn l = map fn l `using` parList rdeepseq
    t3tol (l1, l2, l3) = [l1, l2, l3]
stepForwardAllPar _ _ _ _ = undefined
-}


generateEphemeridesPar
  :: (NFData t, Floating t, Ord t)
  => Integrator t
  -> [ OrbitBody t ] 
  -> CutParams t 
  -> t 
  -> Int 
  -> [ [ OrbitBody t ]]
generateEphemeridesPar integrator bodies grcut dt steps =
  take steps (iterate advance bodies)
  where
    advance bodies' = stepForwardAllPar integrator (f bodies') dt bodies'
    f bs' = integrableForceEq bs' grcut


generateEphemerides
  :: (Floating t, Ord t)
  => Integrator t
  -> [OrbitBody t] 
  -> CutParams t 
  -> t 
  -> Int 
  -> [[OrbitBody t]]
generateEphemerides integrator bodies grcut dt steps =
  take steps (iterate advance bodies)
  where
    advance bodies' = stepForwardAll integrator (f bodies') dt bodies'
    f bs' = integrableForceEq bs' grcut
--generateEphemerides (SingStepIntegrator i) bodies grcut dt steps = undefined
--  --take steps (i (integrableForceEq ))
--generateEphemerides (MultStepIntegrator i) bodies grcut dt steps = undefined
